/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/*
 LD_LIBRARY_PATH=. valgrind ./unnamed
 nslookup -type=mx -port=5353 eder.us 10.0.0.1
 nmap --script dns-fuzz -sU -p 53 10.0.0.1
*/

#include <signal.h>
#include "unnamed_internal.h"

#ifdef _MSC_VER
  #pragma comment(lib, "ws2_32.lib")
#else
  #include <libgen.h>
#endif

#ifdef USE_PTHREADS
  static volatile int running_threads = 0;
  static pthread_mutex_t running_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

static void *cleanup(client_t *client)
{
  char ip_from[IPCHAR_SIZE], ip_to[IPCHAR_SIZE];
  state_t *state = client->state;
  int i;
  for(i = 0; i < client->state->module_count; i++)
  {
    if(client->state->modules[i].collect_garbage != NULL)
    {
      client->state->modules[i].collect_garbage(i, client->state);
    }
  }
  ip_from[0] = ip_to[0] = '\0';
  #ifdef USE_PTHREADS
    if(client->questions[0] != NULL)
    {
      INFO_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: reponse time: %g t=%d %s %s %s", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), microtime() - client->start, running_threads, (char*)client->questions[0]->qname.name.value, type_name(client->questions[0]->qtype), class_name(client->questions[0]->qclass));
    }
    else
    {
      INFO_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: reponse time: %g t=%d", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), microtime() - client->start, running_threads);
    }
  #else
    if(client->questions[0] != NULL)
    {
      INFO_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: reponse time: %g %s %s %s", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), microtime() - client->start, (char*)client->questions[0]->qname.name.value, type_name(client->questions[0]->qtype), class_name(client->questions[0]->qclass));
    }
    else
    {
      INFO_IF(state->trace) (void)lprintf(TRACE, "%s %d => %s %d: reponse time: %g", addr2char(ip_from, &client->from_addr), (int)ntohs(client->from_addr.a.sockaddr_in.sin_port), addr2char(ip_to, &client->to_addr), (int)ntohs(client->to_addr.a.sockaddr_in.sin_port), microtime() - client->start);
    }
  #endif

  (void)free_client(client, 1);
  #ifdef USE_PTHREADS
    (void)pthread_mutex_lock(&running_mutex);
    running_threads--;
    (void)pthread_mutex_unlock(&running_mutex);
  #endif
  return NULL;
}

static void* dns_respond(void *void_client)
{
  ssize_t written;
  client_t *client = (client_t*)void_client;
  state_t *state = client->state;
  int i;
  client->max_udp_size = 512;

  if(client->socket != INVALID_SOCKET) /* tcp */
  {
    int len, length;
    while(client->packet.length < PACKET_LENGTH_SIZE)
    {
      if((len = (int)recv(client->socket, (char*)client->packet.value + client->packet.length, (buf_len)(client->packet.allocated - client->packet.length), 0)) < 0)
      {
        return cleanup(client);
      }
      client->packet.length += (uint32_t)len;
    }

    length = (((int)(client->packet.value[0] << 8)) | (int)client->packet.value[1]) + PACKET_LENGTH_SIZE;
    if(buffer_resize(&client->packet, (uint32_t)length) != 0) return cleanup(client);
    while((int)client->packet.length < length)
    {
      if((len = (int)recv(client->socket, (char*)client->packet.value + client->packet.length, (buf_len)(client->packet.allocated - client->packet.length), 0)) < 0)
      {
        return cleanup(client);
      }
      client->packet.length += (uint32_t)len;
    }
  }

  if(parse_packet(client) != 0) WARNING_IF(state->trace) (void)lprintf(TRACE, "Parse failed");
  if(client->header.flags.opcode != OPCODE_QUERY) WARNING_IF(state->trace) (void)lprintf(TRACE, "HEADER_OPCODE = %x", client->header.flags.opcode);
  client->header.flags.qr = 1;
  client->header.flags.ra = allow_forward(client) & 1;

  if(client->max_udp_size < 512)
  {
    client->max_udp_size = 512;
  }
  else if(client->max_udp_size > 512)
  {
    rr_t *rr = rr_create(NULL, TYPE_OPT, client->max_udp_size, 0, RCODE_NO_ERROR);
    if(rr != NULL)
    {
      (void)add_rr(client, RR_AR, &rr);
      free_rr(&rr);
    }
  }

  for(i = 0; i < (int)client->header.qdcount; i++)
  {
    if(answer_question(client, RR_AN, &client->questions[i]->qname.name, client->questions[i]->qtype, client->questions[i]->qclass, 0) != ANSWER_SATISFIED)
    {
      client->header.flags.rcode = RCODE_NAME_ERROR;
    }
  }

  if(write_packet(client) != 0)
  {
    WARNING_IF(state->trace) (void)lprintf(TRACE, "Failed to write packet to buffer");
    return cleanup(client);
  }

  if(client->socket == INVALID_SOCKET)  /* udp */
  {
    int val;
    if(client->max_udp_size < (uint16_t)client->packet.length - PACKET_LENGTH_SIZE)
    {
      client->packet.length = (uint32_t)client->max_udp_size + PACKET_LENGTH_SIZE;
      client->packet.value[2] |= 2;  /* set TC flag */
    }
    if((client->socket = socket((int)client->to_addr.a.sockaddr.sa_family, SOCK_DGRAM, 0)) == INVALID_SOCKET) return cleanup(client);
    val = 1;
    (void)setsockopt(client->socket, SOL_SOCKET, SO_REUSEADDR, (const char*)&val, (socklen_t)sizeof(val));
    if(bind(client->socket, &client->to_addr.a.sockaddr, (socklen_t)sizeof(client->to_addr)) == -1)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "bind failed: %s", ERRNO_STRING);
      return cleanup(client);
    }
    if((written = sendto(client->socket, (char*)(client->packet.value + PACKET_LENGTH_SIZE), (buf_len)client->packet.length - PACKET_LENGTH_SIZE, 0, &client->from_addr.a.sockaddr, (socklen_t)sizeof(client->from_addr))) != (ssize_t)(client->packet.length - PACKET_LENGTH_SIZE))
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "sendto failed: %s", ERRNO_STRING);
      WARNING_IF(state->trace) (void)lprintf(TRACE, "Failed to write packet to socket %d, %d/%u written", client->socket, (int)written, client->packet.length);
    }
  }
  else  /* tcp */
  {
    if((written = send(client->socket, (char*)client->packet.value, (buf_len)client->packet.length, 0)) != (ssize_t)client->packet.length)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "send failed: %s", ERRNO_STRING);
      WARNING_IF(state->trace) (void)lprintf(TRACE, "Failed to write packet to socket %d, %d/%u written", client->socket, (int)written, client->packet.length);
    }
  }
  return cleanup(client);
}

static volatile sig_atomic_t quit_flag = 0;
static void set_quit_flag(int sig)
{
  #if !defined(WIN32) || defined(__CYGWIN__)
    (void)signal(sig, SIG_IGN);
    quit_flag = 1;
  #elif defined(WIN32)
    if(sig == SIGABRT)
    {
      quit_flag = 1;
    }
  #else
    quit_flag = 1;
  #endif
}

int main(int argc, char *argv[])
{
  address_t servaddr;
  int i, packet_count = 0;
  SOCKET udp_sock, tcp_sock = INVALID_SOCKET;
  state_t _state;
  state_t *state = &_state;
  const char *config = "unnamed.config";
  uint16_t max_udp_size;
  in_port_t server_port;

  #if !defined(WIN32) || defined(__CYGWIN__)
  {
    /* set up sig handler */
    struct sigaction sa;
    memset(&sa, 0, sizeof(sa));
    sa.sa_handler = set_quit_flag;
    (void)sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    if(sigaction(SIGINT, &sa, NULL) == -1)
    {
      (void)printf("sigaction failed: %s", ERRNO_STRING);
      return -1;
    }
  }
  #elif defined(WIN32)
  {
    (void)signal(SIGABRT, set_quit_flag);
  }
  #endif

  for(i = 1; i < argc; i++)
  {
    if(strcmp(argv[i], "-c") == 0 && i + 1 < argc)
    {
      config = argv[i+1];
      i++;
    }
    else if(strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0)
    {
      printf("%s [-c config_file] [option=value] [option=value] ...\n", argv[0]);
      return 0;
    }
  }

  /* attempt to change to unnamed's config directory */
  if(unlikely(chdir("/etc/unnamed/") != 0))
  {
    printf("chdir failed\n");
  }

  memset(state, 0, sizeof(*state));
  if(unnamed_init(state, config, argc, argv) != 0)
  {
    printf("Unable to initialize.\n");
    return -1;
  }

  max_udp_size = (uint16_t)config_get_int(state, "unnamed.max_udp_size", 512);

  server_port = (in_port_t)config_get_int(state, "unnamed.server_port", 53);

  if(config_get_ip(state, "unnamed.server_bind", &servaddr) != 0)
  {
    quit_flag = 1;
  }
  else if(servaddr.a.sockaddr.sa_family == AF_INET)
  {
    servaddr.a.sockaddr_in.sin_port = (in_port_t)htons(server_port);
  }
  else if(servaddr.a.sockaddr.sa_family == AF_INET6)
  {
    servaddr.a.sockaddr_in6.sin6_port = (in_port_t)htons(server_port);
  }
  else
  {
    quit_flag = 1;
  }

  /* set up networking */
  if((udp_sock = socket((int)servaddr.a.sockaddr.sa_family, SOCK_DGRAM, 0)) == INVALID_SOCKET) quit_flag = 1;
  if(udp_sock >= FD_SETSIZE) quit_flag = 1;
  if(quit_flag == 0)
  {
    int val = 1;
    (void)setsockopt(udp_sock, SOL_SOCKET, SO_REUSEADDR, (const char*)&val, (socklen_t)sizeof(val));
    val = 65536;
    (void)setsockopt(udp_sock, SOL_SOCKET, SO_RCVBUF, (const char*)&val, (socklen_t)sizeof(val));
  }
  if(quit_flag == 0)
  {
    if((tcp_sock = socket((int)servaddr.a.sockaddr.sa_family, SOCK_STREAM, 0)) == INVALID_SOCKET) quit_flag = 1;
    if(tcp_sock >= FD_SETSIZE) quit_flag = 1;
    if(quit_flag == 0)
    {
      int val = 1;
      (void)setsockopt(udp_sock, SOL_SOCKET, SO_REUSEADDR, (const char*)&val, (socklen_t)sizeof(val));
    }
  }
  if(quit_flag == 0)
  {
    if(bind(udp_sock, &servaddr.a.sockaddr, (socklen_t)sizeof(servaddr)) == -1)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "bind failed: %s", ERRNO_STRING);
      quit_flag = 1;
    }
  }
  if(quit_flag == 0)
  {
    if(bind(tcp_sock, &servaddr.a.sockaddr, (socklen_t)sizeof(servaddr)) == -1)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "bind failed: %s", ERRNO_STRING);
      quit_flag = 1;
    }
  }
  if(quit_flag == 0)
  {
    if(listen(tcp_sock, 10) == -1)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "listen failed: %s", ERRNO_STRING);
      quit_flag = 1;
    }
  }

  if(quit_flag == 0) (void)drop_priv(state);

  while(quit_flag == 0)
  {
    int length;
    fd_set rfds;
    #ifdef USE_PTHREADS
      pthread_t child;
    #endif

    client_t *client = (client_t*)calloc(1, sizeof(client_t));
    socklen_t addr_len = (socklen_t)sizeof(client->from_addr);
    if(client == NULL)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "calloc failed: %s", ERRNO_STRING);
      break;
    }
    if(init_client(state, client, max_udp_size) != 0)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "buffer_resize failed");
      (void)free_client(client, 1);
      break;
    }

    FD_ZERO(&rfds);
    FD_SET((unsigned int)udp_sock, &rfds);
    if(tcp_sock != INVALID_SOCKET) { FD_SET((unsigned int)tcp_sock, &rfds); }
    if(select((int)((udp_sock > tcp_sock) ? udp_sock : tcp_sock) + 1, &rfds, NULL, NULL, NULL) == -1)
    {
      if(quit_flag == 0) ERROR_IF(state->trace) (void)lprintf(TRACE, "select failed: %s", ERRNO_STRING);
      (void)free_client(client, 1);
      continue;
    }
    if(FD_ISSET((unsigned int)udp_sock, &rfds))
    {
      if((length = (int)recvfrom(udp_sock, (char*)client->packet.value + PACKET_LENGTH_SIZE, (buf_len)(client->packet.allocated - PACKET_LENGTH_SIZE), 0, (struct sockaddr*)&client->from_addr, &addr_len)) < 12)
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "recvfrom failed: %s", ERRNO_STRING);
        packet_count++;
        (void)free_client(client, 1);
        continue;
      }
      client->packet.value[0] = (uint8_t)(((unsigned int)length & 0xff00) >> 8);
      client->packet.value[1] = (uint8_t)(length & 0xff);
    }
    else if(tcp_sock != INVALID_SOCKET && FD_ISSET((unsigned int)tcp_sock, &rfds))
    {
      client->socket = accept(tcp_sock, (struct sockaddr*)&client->from_addr, &addr_len);
      if(client->socket == INVALID_SOCKET)
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "accept failed: %s", ERRNO_STRING);
        (void)free_client(client, 1);
        continue;
      }
      if((length = (int)recv(client->socket, (char*)client->packet.value, (buf_len)client->packet.allocated, 0)) < 2)
      {
        if(errno != 0) ERROR_IF(state->trace) (void)lprintf(TRACE, "recv failed: %s", ERRNO_STRING);
        packet_count++;
        (void)free_client(client, 1);
        continue;
      }
    }
    else
    {
      continue;
    }

    client->start = microtime();
    packet_count++;
    client->packet.length = (uint32_t)length;

    (void)my_addr(&client->to_addr, &client->from_addr, server_port);

    #ifdef USE_PTHREADS
      (void)pthread_mutex_lock(&running_mutex);
      running_threads++;
      (void)pthread_mutex_unlock(&running_mutex);
      if(pthread_create(&child, NULL, dns_respond, client) != 0)
      {
        (void)free_client(client, 1);
        ERROR_IF(state->trace) (void)lprintf(TRACE, "pthread_create failed: %s", ERRNO_STRING);
        break;
      }
      (void)pthread_detach(child);
    #else
      (void)dns_respond(client);
    #endif
  }

  DEBUG_IF(state->trace) (void)lprintf(TRACE, "cleanup ...");
  INFO_IF(state->trace) (void)lprintf(TRACE, "%d packets processed", packet_count);
  if(udp_sock != INVALID_SOCKET) (void)closesocket(udp_sock);
  if(tcp_sock != INVALID_SOCKET) (void)closesocket(tcp_sock);
  #ifdef USE_PTHREADS
    while(running_threads > 0) (void)usleep(250000);
  #endif
  unnamed_term(state);
  config_free(state);
  #ifdef USE_PTHREADS
    pthread_exit(NULL);
  #endif

  return -1;
}
