/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdarg.h>
#include <fcntl.h>
#if !defined(WIN32) || defined(__CYGWIN__)
  #include <pwd.h>
#endif
#include "unnamed_internal.h"

#ifdef _MSC_VER
static const char *dlerror(void)
{
  DWORD dw = GetLastError();
  if(dw != ERROR_SUCCESS)
  {
    static char buffer[4096];
    LPVOID lpMsgBuf;
    unsigned int i;
    DWORD length = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, dw,
                                 MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&lpMsgBuf, 0, NULL);
    for(i = 0; i < length; i++) buffer[i] = ((char*)lpMsgBuf)[i];
    buffer[i] = '\0';
    LocalFree(lpMsgBuf);
    return buffer;
  }
  return NULL;
}
#endif

static int load_module(state_t *state, const char *name, const char *module_dir)
{
  const char *error;
  buffer_t buf;
  int i;
  module_t *module;
  if(unlikely(state == NULL || name == NULL || module_dir == NULL)) return -1;

  for(i = 0; i < state->module_count; i++)
  {
    if(strcmp(state->modules[i].so, name) == 0)
    {
      memcpy(&state->modules[state->module_count], &state->modules[i], sizeof(state->modules[0]));
      if(likely(state->modules[state->module_count].init != NULL))
      {
        if(unlikely(state->modules[state->module_count].init(state->module_count, state) != 0))
        {
          ERROR_IF(state->trace) (void)lprintf(TRACE, "Error: %s failed to initialize", name);
          return -1;
        }
      }
      state->modules[state->module_count].so = strdup(state->modules[i].so);
      state->module_handle[state->module_count] = NULL;
      state->module_count++;
      return 0;
    }
  }

  buffer_init(&buf);
  if(buffer_set(&buf, "%s" DS "unnamed_%s." SO_EXT, module_dir, name) != 0) return -1;

  if(unlikely((state->module_handle[state->module_count] = dlopen((char*)buf.value, RTLD_NOW)) == NULL))
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "Unable to load %s from %s: %s", name, (char*)buf.value, dlerror());
    buffer_free(&buf);
    return -1;
  }
  if(unlikely((error = dlerror()) != NULL))
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "dlopen %s: %s", name, error);
  }
  buffer_free(&buf);

  if(unlikely((module = (module_t*)dlsym(state->module_handle[state->module_count], "unnamed_module")) == NULL))
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "Unable to load unnamed_module in %s: %s", name, dlerror());
    (void)dlclose(state->module_handle[state->module_count]);
    return -1;
  }
  if(unlikely((error = dlerror()) != NULL))
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "dlopen %s: %s", name, error);
  }

  memcpy(&state->modules[state->module_count], module, sizeof(state->modules[0]));

  if(unlikely(state->modules[state->module_count].answer == NULL))
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "Error: %s does not answer", name);
    (void)dlclose(state->module_handle[state->module_count]);
    return -1;
  }

  if(likely(state->modules[state->module_count].init != NULL))
  {
    if(unlikely(state->modules[state->module_count].init(state->module_count, state) != 0))
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "Error: %s failed to initialize", name);
      (void)dlclose(state->module_handle[state->module_count]);
      return -1;
    }
  }

  state->modules[state->module_count].so = strdup(name);
  state->module_count++;
  return 0;
}

#ifdef _MSC_VER
struct timezone
{
  int tz_minuteswest;     /* minutes west of Greenwich */
  int tz_dsttime;         /* type of DST correction */
};
static int gettimeofday(struct timeval *tv, struct timezone *tz)
{
  if(tv != NULL)
  {
    FILETIME ft;
    __int64 tmp64;
    memset(&ft, 0, sizeof(ft));
    GetSystemTimeAsFileTime(&ft);
    tmp64 = (((__int64)ft.dwHighDateTime << 32) | (__int64)ft.dwLowDateTime) / 10LL - 11644473600000000LL;
    tv->tv_sec = (__int32)(tmp64 / 1000000LL);
    tv->tv_usec = tmp64 % 1000000;
  }
  if(tz != NULL)
  {
    TIME_ZONE_INFORMATION tz_winapi;
    memset(&tz_winapi, 0, sizeof(tz_winapi));
    if(GetTimeZoneInformation(&tz_winapi) == 2)
    {
      tz->tz_dsttime = 1;
      tz->tz_minuteswest = tz_winapi.Bias + tz_winapi.DaylightBias;
    }
    else
    {
      tz->tz_dsttime = 0;
      tz->tz_minuteswest = tz_winapi.Bias;
    }
  }
  return 0;
}
#endif

int unnamed_init(state_t *state, const char *config, int argc, char *argv[])
{
  int i, count;
  buffer_t buffer;
  const char *lprintf_filename;
  struct timeval tv;
  unsigned int seed = 0;

  #ifdef WIN32
    WSADATA wsaData;
    WORD wVersionRequested = MAKEWORD(2, 2);
    WSAStartup(wVersionRequested, &wsaData);
  #endif

  if(unlikely(state == NULL)) return -1;
  if(unlikely(config == NULL)) return -1;

  (void)gettimeofday(&tv, NULL);
  #if !defined(WIN32) || defined(__CYGWIN__)
    if((state->random_fd = open("/dev/urandom", O_RDONLY)) >= 0)
    {
      if(read(state->random_fd, &seed, sizeof(seed)) <= 0)
      {
        seed = (unsigned int)(long)&seed;
      }
    }
  #else
    seed = (unsigned int)(long)&seed;
  #endif
  srandom(seed ^ (unsigned int)tv.tv_usec);

  /* read config file */
  buffer_init(&state->config.name);
  buffer_init(&state->config.value);
  state->config.child_count = 0;
  state->config.children = NULL;

  if(unlikely(config_read(config, &state->config) != 0))
  {
    unnamed_term(state);
    return -1;
  }

  for(i = 1; i < argc; i++)
  {
    if(argv[i][0] == '-')
    {
      i++;
    }
    else
    {
      char *value = strchr(argv[i], '=');
      if(value != NULL)
      {
        *value++ = '\0';
        (void)config_set(state, &state->config, argv[i], value);
      }
      else
      {
        (void)config_set(state, &state->config, argv[i], NULL);
      }
    }
  }

  state->trace = config_get_int(state, "unnamed.trace", TRACE_LEVEL_ERROR);

  state->buffer_free = buffer_free;
  state->buffer_resize = buffer_resize;
  state->buffer_copy = buffer_copy;
  state->buffer_set = buffer_set;
  state->buffer_cmp = buffer_cmp;
  state->buffer_put_raw = buffer_put_raw;
  state->addr2char = addr2char;
  state->char2addr = char2addr;
  state->addr_cmp = addr_cmp;
  state->is_in_range = is_in_range;
  state->addr_prefix = addr_prefix;
  state->addr6_prefix = addr6_prefix;
  state->is_subdomain = is_subdomain;
  state->ptr_name = ptr_name;
  state->config_get_const = config_get_const;
  state->config_get = config_get;
  state->config_get_lower = config_get_lower;
  state->config_get_int = config_get_int;
  state->config_get_double = config_get_double;
  state->config_get_ip = config_get_ip;
  state->config_get_ip_range = config_get_ip_range;
  state->config_free_node = config_free_node;
  state->config_maxn = config_maxn;
  state->is_ip_black = is_ip_black;
  state->is_name_white = is_name_white;
  state->is_name_black = is_name_black;
  state->question_create = question_create;
  state->free_rr = free_rr;
  state->add_rr = add_rr;
  state->rr_create = rr_create;
  state->parse_packet = parse_packet;
  state->write_packet = write_packet;
  state->init_client = init_client;
  state->free_client = free_client;
  state->microtime = microtime;
  state->current_serial = current_serial;
  state->type_name = type_name;
  state->class_name = class_name;
  state->answer_type_name = answer_type_name;
  state->rcode_name = rcode_name;
  state->answer_question = answer_question;
  state->match = unnamed_match;
  state->random_id = random_id;
  state->lprintf = lprintf;

  state->min_ttl = (uint32_t)config_get_int(state, "unnamed.min_ttl", 0);

  lprintf_filename = config_get_const(state, "unnamed.log");
  (void)fclose(stdin);
  (void)fclose(stderr);
  state->lprintf_fp = stdout;
  if(lprintf_filename != NULL)
  {
    FILE *fp;
    if(unlikely((fp = fopen(lprintf_filename, "a")) != NULL))
    {
      state->lprintf_fp = fp;
      (void)fclose(stdout);
    }
  }

  if(state->trace >= TRACE_LEVEL_INFO) config_dump(state);

  if(unlikely(black_white_init(state) != 0))
  {
    unnamed_term(state);
    return -1;
  }

  buffer_init(&buffer);
  count = config_maxn(state, "unnamed.module");
  if(unlikely((state->modules = (module_t*)calloc((size_t)count, sizeof(module_t))) == NULL))
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "calloc failed");
    unnamed_term(state);
    return -1;
  }
  if(unlikely((state->module_data = (void**)calloc((size_t)count, sizeof(void*))) == NULL))
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "calloc failed");
    unnamed_term(state);
    return -1;
  }
  if(unlikely((state->module_handle = (void**)calloc((size_t)count, sizeof(void*))) == NULL))
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "calloc failed");
    unnamed_term(state);
    return -1;
  }
  else
  {
    const char *module_dir = config_get_const(state, "unnamed.module_dir");
    if(module_dir == NULL) module_dir = MODULE_DIR;
    for(i = 0; i < count; i++)
    {
      const char *module_name;
      if(unlikely(buffer_set(&buffer, "unnamed.module.%d", i + 1) != 0))
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "buffer_set failed");
        unnamed_term(state);
        return -1;
      }
      if(unlikely((module_name = config_get_const(state, (char*)buffer.value)) == NULL))
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "config_get_const failed on %s", (char*)buffer.value);
        unnamed_term(state);
        return -1;
      }
      if(unlikely(load_module(state, module_name, module_dir) != 0))
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "load_module failed on %s", module_name);
        unnamed_term(state);
        return -1;
      }
    }
  }
  count = config_maxn(state, "unnamed.allow_forward");
  if(unlikely((state->allow_forwards = (ip_range_t*)calloc((size_t)count, sizeof(state->allow_forwards[0]))) == NULL))
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "calloc failed");
    unnamed_term(state);
    return -1;
  }
  for(state->allow_forwards_count = 0; state->allow_forwards_count < count; state->allow_forwards_count++)
  {
    if(unlikely(buffer_set(&buffer, "unnamed.allow_forward.%d", state->allow_forwards_count + 1) != 0))
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "buffer_set failed");
      unnamed_term(state);
      return -1;
    }
    memset(&state->allow_forwards[state->allow_forwards_count], 0, sizeof(state->allow_forwards[0]));
    if(unlikely(config_get_ip_range(state, (char*)buffer.value, &state->allow_forwards[state->allow_forwards_count]) != 0))
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "config_get_ip_range failed on %s", (char*)buffer.value);
      unnamed_term(state);
      return -1;
    }
  }
  buffer_free(&buffer);

  return 0;
}

void unnamed_term(state_t *state)
{
  int i;
  if(unlikely(state != NULL))
  {
    for(i = state->module_count - 1; i >= 0; i--)
    {
      if(likely(state->modules[i].term != NULL)) state->modules[i].term(i, state);
      if(likely(state->modules[i].so != NULL)) free((void*)state->modules[i].so);
      if(likely(state->module_handle[i] != NULL)) (void)dlclose(state->module_handle[i]);
    }
    black_white_term(state);
    if(likely(state->allow_forwards != NULL))
    {
      free(state->allow_forwards);
      state->allow_forwards_count = 0;
      state->allow_forwards = NULL;
    }
    if(likely(state->modules != NULL))
    {
      free(state->modules);
      state->modules = NULL;
    }
    if(likely(state->module_data != NULL))
    {
      free(state->module_data);
      state->module_data = NULL;
    }
    if(likely(state->module_handle != NULL))
    {
      free(state->module_handle);
      state->module_handle = NULL;
    }
    state->module_count = 0;
    if(state->lprintf_fp != NULL) (void)fclose(state->lprintf_fp);
    #if !defined(WIN32) || defined(__CYGWIN__)
      if(state->random_fd >= 0) (void)close(state->random_fd);
    #endif
  }
  #ifdef WIN32
    WSACleanup();
  #endif
}

question_t* question_create(const buffer_t *name, uint16_t type, uint16_t dclass)
{
  question_t *q;
  if(unlikely((q = (question_t*)calloc(1, sizeof(question_t))) == NULL)) return NULL;
  buffer_init(&q->qname.name);
  if(likely(name != NULL))
  {
    if(unlikely(buffer_copy(&q->qname.name, name) != 0))
    {
      free(q);
      return NULL;
    }
  }
  q->qtype = type;
  q->qclass = dclass;
  return q;
}

void free_rr(rr_t **rr)
{
  if(unlikely(rr == NULL || *rr == NULL)) return;
  buffer_free(&(*rr)->name.name);
  if(likely((*rr)->rclass == CLASS_IN))
  {
    switch((*rr)->rtype)
    {
      case TYPE_CNAME:
      {
        buffer_free(&(*rr)->p.cname.name);
      } break;
      case TYPE_MX:
      {
        buffer_free(&(*rr)->p.mx.exchange.name);
      } break;
      case TYPE_NS:
      {
        buffer_free(&(*rr)->p.ns.name);
      } break;
      case TYPE_PTR:
      {
        buffer_free(&(*rr)->p.ptr.name);
      } break;
      case TYPE_SOA:
      {
        buffer_free(&(*rr)->p.soa.mname.name);
        buffer_free(&(*rr)->p.soa.rname.name);
      } break;
      case TYPE_SPF:
      {
        buffer_free(&(*rr)->p.spf);
      } break;
      case TYPE_TXT:
      {
        buffer_free(&(*rr)->p.txt);
      } break;
      default: break;
    }
  }
  else if(likely((*rr)->rclass == CLASS_CH && (*rr)->rtype == TYPE_TXT)) buffer_free(&(*rr)->p.txt);
  free(*rr);
  *rr = NULL;
}

rr_t* rr_create(const buffer_t *name, uint16_t type, uint16_t dclass, uint32_t expires, uint8_t status)
{
  rr_t *rr;
  if(unlikely((rr = (rr_t*)calloc(1, sizeof(rr_t))) == NULL)) return NULL;
  memset(&rr->p, 0, sizeof(rr->p));
  buffer_init(&rr->name.name);
  rr->next = NULL;
  rr->rclass = dclass;
  rr->rtype = type;
  rr->expires = expires;
  rr->status = status;
  if(likely(name != NULL))
  {
    if(unlikely(buffer_copy(&rr->name.name, name) != 0)) free_rr(&rr);
  }
  return rr;
}

int add_rr(client_t *client, rr_section_t rr_section, rr_t **rr)
{
  if(unlikely(client == NULL || rr == NULL)) return -1;
  switch(rr_section)
  {
    case RR_AN:
    {
      if(likely(client->header.ancount + 1 < MAX_RRS))
      {
        client->answers[client->header.ancount++] = *rr;
        *rr = NULL;
      }
    } break;
    case RR_NS:
    {
      if(likely(client->header.nscount + 1 < MAX_RRS))
      {
        client->name_servers[client->header.nscount++] = *rr;
        *rr = NULL;
      }
    } break;
    case RR_AR:
    {
      if(likely(client->header.arcount + 1 < MAX_RRS))
      {
        client->additional_records[client->header.arcount++] = *rr;
        *rr = NULL;
      }
    } break;
    default: return -1;
  }
  return 0;
}

static int get_header(client_t *client)
{
  if(unlikely(client == NULL)) return -1;
  if(unlikely(buffer_get16(&client->packet, &client->header.id) != 0)) return -1;
  if(unlikely(buffer_get_raw(&client->packet, (uint8_t*)&client->header.flags, 2) != 0)) return -1;
  if(unlikely(buffer_get16(&client->packet, &client->header.qdcount) != 0)) return -1;
  if(unlikely(buffer_get16(&client->packet, &client->header.ancount) != 0)) return -1;
  if(unlikely(buffer_get16(&client->packet, &client->header.nscount) != 0)) return -1;
  if(unlikely(buffer_get16(&client->packet, &client->header.arcount) != 0)) return -1;
  if(unlikely(client->header.qdcount > MAX_QUESTIONS))
  {
    client->header.qdcount = MAX_QUESTIONS;
    client->header.ancount = 0;
    client->header.nscount = 0;
    client->header.arcount = 0;
    return 0;
  }
  if(unlikely(client->header.ancount > MAX_RRS))
  {
    client->header.ancount = MAX_RRS;
    client->header.nscount = 0;
    client->header.arcount = 0;
    return 0;
  }
  if(unlikely(client->header.nscount > MAX_RRS))
  {
    client->header.nscount = MAX_RRS;
    client->header.arcount = 0;
    return 0;
  }
  if(unlikely(client->header.arcount > MAX_RRS))
  {
    client->header.arcount = MAX_RRS;
    return 0;
  }
  return 0;
}

static int append_labels(const buffer_t *data, uint32_t *offset, buffer_t *name, int depth)
{
  int ret = 0;
  uint32_t len;
  if(unlikely(data == NULL || offset == NULL || name == NULL)) return -1;
  for(;;)
  {
    if(unlikely(*offset >= data->length)) break;
    if((data->value[*offset] & 0xc0) != 0)
    {
      uint32_t o;
      if(depth == 0)
      {
        ret = -1;
        break;
      }
      if(*offset + 1 >= data->length) return -1;
      o = ((((uint32_t)data->value[*offset] << 8) | (uint32_t)data->value[*offset + 1]) & 0x3fff) + PACKET_LENGTH_SIZE;
      *offset += 2;
      return append_labels(data, &o, name, depth - 1);
    }
    len = (uint32_t)data->value[(*offset)++];
    if(unlikely(len == 0 || *offset + len >= data->length)) break;
    if(unlikely(buffer_resize(name, name->length + 1 + len) != 0)) return -1;
    if(likely(name->length != 0)) name->value[name->length++] = (uint8_t)'.';
    if(unlikely(buffer_put_raw(name, data->value + *offset, len) != 0)) return -1;
    *offset += len;
  }
  if(likely(buffer_resize(name, name->length + 1) == 0)) name->value[name->length] = 0;
  return ret;
}

static void free_question(client_t *client, int i)
{
  if(likely(client != NULL && client->questions[i] != NULL))
  {
    buffer_free(&client->questions[i]->qname.name);
    free(client->questions[i]);
    client->questions[i] = NULL;
  }
}

static int get_name(buffer_t *data, name_t *name)
{
  int ret;
  if(unlikely(data == NULL || name == NULL)) return -1;
  buffer_init(&name->name);
  ret = append_labels(data, &data->offset, &name->name, MAX_LABELS);
  name->name.value[name->name.length] = 0;
  return ret;
}

static int get_question(client_t *client, int i)
{
  if(unlikely(client == NULL)) return -1;
  if(unlikely((client->questions[i] = question_create(NULL, 0, 0)) == NULL)) return -1;
  if(likely(get_name(&client->packet, &client->questions[i]->qname) == 0))
  {
    if(likely(client->questions[i]->qname.name.length > 0))
    {
      if(likely(buffer_get16(&client->packet, &client->questions[i]->qtype) == 0))
      {
        if(likely(buffer_get16(&client->packet, &client->questions[i]->qclass) == 0)) return 0;
      }
    }
  }
  free_question(client, i);
  return -1;
}

static void free_client_rr(client_t *client, rr_section_t rr_section, int i)
{
  if(unlikely(client == NULL)) return;
  switch(rr_section)
  {
    case RR_AN: free_rr(&client->answers[i]); client->answers[i] = NULL; break;
    case RR_NS: free_rr(&client->name_servers[i]); client->name_servers[i] = NULL; break;
    case RR_AR: free_rr(&client->additional_records[i]); client->additional_records[i] = NULL; break;
  }
}

static int get_rr(client_t *client, rr_section_t rr_section, int i)
{
  rr_t *rr;
  uint32_t ttl = 0, ttl_raw = 0;
  uint32_t rdata_offset;
  uint16_t rdata_length = 0;
  state_t *state;

  #define CLEANUP_FAIL()  { WARNING_IF(state->trace) (void)lprintf(TRACE, "failed to read RR"); free_rr(&rr); return -1; }

  if(unlikely(client == NULL)) return -1;
  state = client->state;

  free_client_rr(client, rr_section, i);

  if(unlikely((rr = rr_create(NULL, 0, 0, 0, RCODE_NO_ERROR)) == NULL)) return -1;
  if(unlikely(get_name(&client->packet, &rr->name) != 0)) CLEANUP_FAIL();
  if(unlikely(buffer_get16(&client->packet, &rr->rtype) != 0)) CLEANUP_FAIL();
  if(unlikely(buffer_get16(&client->packet, &rr->rclass) != 0)) CLEANUP_FAIL();
  if(unlikely(buffer_get32(&client->packet, &ttl_raw) != 0)) CLEANUP_FAIL();
  if((ttl_raw & 0x80000000) != 0)
  {
    ttl = 0; /* rfc 2181, section 8 */
  }
  else
  {
    ttl = ttl_raw;
  }
  if(ttl < state->min_ttl) ttl = state->min_ttl;
  rr->expires = (uint32_t)time(NULL) + ttl;

  if(client->packet.offset == client->packet.length)
  {
    rdata_length = 0;
  }
  else
  {
    if(unlikely(buffer_get16(&client->packet, &rdata_length) != 0)) CLEANUP_FAIL();
  }
  rdata_offset = client->packet.offset;
  if(unlikely(client->packet.offset + (unsigned int)rdata_length > client->packet.length))
  {
    WARNING_IF(state->trace) (void)lprintf(TRACE, "client->packet.offset=%d rdata_length=%d client->packet.length=%d", client->packet.offset, rdata_length, client->packet.length);
    CLEANUP_FAIL();
  }

  if(rr->rtype == TYPE_OPT)
  {
    rr->expires = (uint32_t)time(NULL) - 1;
    client->max_udp_size = rr->rclass;
  }
  else
  {
    if(likely(rr->rclass == CLASS_IN))
    {
      switch(rr->rtype)
      {
        case TYPE_A:
        {
          if(unlikely(buffer_get_addr(&client->packet, &rr->p.address) != 0)) CLEANUP_FAIL();
        } break;
        case TYPE_AAAA:
        {
          if(unlikely(buffer_get_addr6(&client->packet, &rr->p.address) != 0)) CLEANUP_FAIL();
        } break;
        case TYPE_CNAME:
        {
          if(unlikely(get_name(&client->packet, &rr->p.cname) != 0)) CLEANUP_FAIL();
        } break;
        case TYPE_MX:
        {
          if(unlikely(buffer_get16(&client->packet, &rr->p.mx.preference) != 0)) CLEANUP_FAIL();
          if(unlikely(get_name(&client->packet, &rr->p.mx.exchange) != 0)) CLEANUP_FAIL();
        } break;
        case TYPE_NS:
        {
          if(unlikely(get_name(&client->packet, &rr->p.ns) != 0)) CLEANUP_FAIL();
        } break;
        case TYPE_PTR:
        {
          if(unlikely(get_name(&client->packet, &rr->p.ptr) != 0)) CLEANUP_FAIL();
        } break;
        case TYPE_SOA:
        {
          if(unlikely(get_name(&client->packet, &rr->p.soa.mname) != 0)) CLEANUP_FAIL();
          if(unlikely(get_name(&client->packet, &rr->p.soa.rname) != 0)) CLEANUP_FAIL();
          if(unlikely(buffer_get32(&client->packet, &rr->p.soa.serial) != 0)) CLEANUP_FAIL();
          if(unlikely(buffer_get32(&client->packet, &rr->p.soa.refresh) != 0)) CLEANUP_FAIL();
          if(unlikely(buffer_get32(&client->packet, &rr->p.soa.retry) != 0)) CLEANUP_FAIL();
          if(unlikely(buffer_get32(&client->packet, &rr->p.soa.expire) != 0)) CLEANUP_FAIL();
          if(unlikely(buffer_get32(&client->packet, &rr->p.soa.minimum) != 0)) CLEANUP_FAIL();
        } break;
        case TYPE_SPF:
        {
          if(unlikely(buffer_resize(&rr->p.spf, rdata_length) != 0)) CLEANUP_FAIL();
          if(unlikely(buffer_get_raw(&client->packet, rr->p.spf.value, rdata_length) != 0)) CLEANUP_FAIL();
          rr->p.spf.length = rdata_length;
        } break;
        case TYPE_TXT:
        {
          if(unlikely(buffer_resize(&rr->p.txt, rdata_length) != 0)) CLEANUP_FAIL();
          if(unlikely(buffer_get_raw(&client->packet, rr->p.txt.value, rdata_length) != 0)) CLEANUP_FAIL();
          rr->p.txt.length = rdata_length;
        } break;
        default:
        {
          WARNING_IF(state->trace) (void)lprintf(TRACE, "type = %s (%u)", type_name(rr->rtype), rr->rtype);
        } break;
      }
    }
  }
  client->packet.offset = rdata_offset + (unsigned int)rdata_length;
  switch(rr_section)
  {
    case RR_AN: client->answers[i] = rr; break;
    case RR_NS: client->name_servers[i] = rr; break;
    case RR_AR: client->additional_records[i] = rr; break;
    default: CLEANUP_FAIL();
  }
  return 0;
  #undef CLEANUP_FAIL
}

int parse_packet(client_t *client)
{
  state_t *state;
  int i, ret;
  if(unlikely(client == NULL)) return -1;
  state = client->state;
  for(i = 0; i < (int)client->header.qdcount; i++) free_question(client, i);
  for(i = 0; i < (int)client->header.ancount; i++) free_client_rr(client, RR_AN, i);
  for(i = 0; i < (int)client->header.nscount; i++) free_client_rr(client, RR_NS, i);
  for(i = 0; i < (int)client->header.arcount; i++) free_client_rr(client, RR_AR, i);
  client->packet.offset = PACKET_LENGTH_SIZE;
  ret = get_header(client);
  for(i = 0; ret == 0 && i < (int)client->header.qdcount; i++) ret = get_question(client, i);
  if(unlikely(ret != 0))
  {
    WARNING_IF(state->trace) (void)lprintf(TRACE, "Parse failed on question %d", i - 1);
    client->header.qdcount = (uint16_t)(i - 1);
    client->header.ancount = 0;
    client->header.nscount = 0;
    client->header.arcount = 0;
  }
  else
  {
    for(i = 0; ret == 0 && i < (int)client->header.ancount; i++) ret = get_rr(client, RR_AN, i);
    if(unlikely(ret != 0))
    {
      WARNING_IF(state->trace) (void)lprintf(TRACE, "Parse failed on answer %d", i - 1);
      client->header.ancount = (uint16_t)(i - 1);
      client->header.nscount = 0;
      client->header.arcount = 0;
    }
    else
    {
      for(i = 0; ret == 0 && i < (int)client->header.nscount; i++) ret = get_rr(client, RR_NS, i);
      if(unlikely(ret != 0))
      {
        WARNING_IF(state->trace) (void)lprintf(TRACE, "Parse failed on name server %d", i - 1);
        client->header.nscount = (uint16_t)(i - 1);
        client->header.arcount = 0;
      }
      else
      {
        for(i = 0; ret == 0 && i < (int)client->header.arcount; i++) ret = get_rr(client, RR_AR, i);
        if(unlikely(ret != 0))
        {
          WARNING_IF(state->trace) (void)lprintf(TRACE, "Parse failed on additional record %d", i - 1);
          client->header.arcount = (uint16_t)(i - 1);
        }
      }
    }
  }
  return ret;
}

static int put_header(client_t *client)
{
  if(unlikely(client == NULL)) return -1;
  if(unlikely(buffer_put16(&client->packet, client->header.id) != 0)) return -1;
  if(unlikely(buffer_put_raw(&client->packet, (uint8_t*)&client->header.flags, 2) != 0)) return -1;
  if(unlikely(buffer_put16(&client->packet, client->header.qdcount) != 0)) return -1;
  if(unlikely(buffer_put16(&client->packet, client->header.ancount) != 0)) return -1;
  if(unlikely(buffer_put16(&client->packet, client->header.nscount) != 0)) return -1;
  if(unlikely(buffer_put16(&client->packet, client->header.arcount) != 0)) return -1;
  return 0;
}

static int find_name_rr(rr_t *rr, buffer_t *name)
{
  if(unlikely(rr == NULL || name == NULL)) return -1;
  if(rr->name.offset == 0) return 0;
  if(buffer_cmp(&rr->name.name, name) == 0) return rr->name.offset;

  if(rr->rclass == CLASS_IN)
  {
    switch(rr->rtype)
    {
      case TYPE_CNAME:
      {
        if(rr->p.cname.offset == 0) return 0;
        if(buffer_cmp(&rr->p.cname.name, name) == 0) return rr->p.cname.offset;
      } break;
      case TYPE_MX:
      {
        if(rr->p.mx.exchange.offset == 0) return 0;
        if(buffer_cmp(&rr->p.mx.exchange.name, name) == 0) return rr->p.mx.exchange.offset;
      } break;
      case TYPE_NS:
      {
        if(rr->p.ns.offset == 0) return 0;
        if(buffer_cmp(&rr->p.ns.name, name) == 0) return rr->p.ns.offset;
      } break;
      case TYPE_PTR:
      {
        if(rr->p.ptr.offset == 0) return 0;
        if(buffer_cmp(&rr->p.ptr.name, name) == 0) return rr->p.ptr.offset;
      } break;
      case TYPE_SOA:
      {
        if(rr->p.soa.mname.offset == 0) return 0;
        if(buffer_cmp(&rr->p.soa.mname.name, name) == 0) return rr->p.soa.mname.offset;
        if(rr->p.soa.rname.offset == 0) return 0;
        if(buffer_cmp(&rr->p.soa.rname.name, name) == 0) return rr->p.soa.rname.offset;
      } break;
      default: break;
    }
  }
  return -1;
}

__attribute__((pure)) static int find_name(client_t *client, buffer_t *name)
{
  int i, ret;
  if(unlikely(client == NULL || name == NULL)) return -1;
  for(i = 0; i < (int)client->header.qdcount; i++)
  {
    if(client->questions[i]->qname.offset == 0) return 0;
    if(buffer_cmp(&client->questions[i]->qname.name, name) == 0) return client->questions[i]->qname.offset;
  }
  for(i = 0; i < (int)client->header.ancount; i++) if((ret = find_name_rr(client->answers[i], name)) >=  0) return ret;
  for(i = 0; i < (int)client->header.nscount; i++) if((ret = find_name_rr(client->name_servers[i], name)) >= 0) return ret;
  for(i = 0; i < (int)client->header.arcount; i++) if((ret = find_name_rr(client->additional_records[i], name)) >= 0) return ret;
  return 0;
}

static int put_name(client_t *client, buffer_t *data, name_t *name, int record_offset_as)
{
  uint32_t i, b=0;
  unsigned int pointer;
  if(unlikely(client == NULL || name == NULL)) return -1;
  pointer = (unsigned int)find_name(client, &name->name);
  name->offset = record_offset_as - PACKET_LENGTH_SIZE;
  if(pointer > 0 && pointer <= 0x3fff)
  {
    if(unlikely(buffer_resize(data, data->length + 2) != 0)) return -1;
    data->value[data->length++] = (uint8_t)(0xc0 | (pointer >> 8));
    data->value[data->length++] = (uint8_t)(pointer & 0xff);
    return 0;
  }
  else
  {
    if(unlikely(buffer_resize(data, data->length + 2 + name->name.length) != 0)) return -1;
    for(i = 0; i < name->name.length; i++)
    {
      if(name->name.value[i] == (uint8_t)'.')
      {
        if(i - b == 0) break;
        data->value[data->length++] = (uint8_t)(i - b);
        memcpy(&data->value[data->length], &name->name.value[b], (size_t)(i - b));
        data->length += i - b;
        b = i + 1;
      }
    }
    if(i - b > 0)
    {
      data->value[data->length++] = (uint8_t)(i - b);
      memcpy(&data->value[data->length], &name->name.value[b], (size_t)(i - b));
      data->length += i - b;
    }
    data->value[data->length++] = 0;
  }
  return 0;
}

static int put_question(client_t *client, int i)
{
  if(unlikely(client == NULL)) return -1;
  if(unlikely(put_name(client, &client->packet, &client->questions[i]->qname, (int)client->packet.length) != 0)) return -1;
  if(unlikely(buffer_put16(&client->packet, client->questions[i]->qtype)  != 0)) return -1;
  if(unlikely(buffer_put16(&client->packet, client->questions[i]->qclass) != 0)) return -1;
  return 0;
}

static int put_rr(client_t *client, rr_section_t rr_section, int i)
{
  uint32_t ttl, now = (uint32_t)time(NULL);
  rr_t *rr = NULL;
  buffer_t rdata;
  int ret;

  if(unlikely(client == NULL)) return -1;
  switch(rr_section)
  {
    case RR_AN: rr = client->answers[i]; break;
    case RR_NS: rr = client->name_servers[i]; break;
    case RR_AR: rr = client->additional_records[i]; break;
  }
  if(unlikely(rr == NULL)) return -1;

  if(unlikely(put_name(client, &client->packet, &rr->name, (int)client->packet.length) != 0)) return -1;
  if(unlikely(buffer_put16(&client->packet, rr->rtype) != 0)) return -1;
  if(unlikely(buffer_put16(&client->packet, rr->rclass) != 0)) return -1;
  if(rr->rtype == TYPE_OPT)
  {
    ttl = rr->expires;
  }
  else
  {
    ttl = (now >= rr->expires) ? 0 : rr->expires - now;
  }
  if(unlikely(buffer_put32(&client->packet, ttl) != 0)) return -1;

  buffer_init(&rdata);

  if(likely(rr->rclass == CLASS_IN))
  {
    switch(rr->rtype)
    {
      case TYPE_A:
      {
        if(unlikely(buffer_put_addr(&rdata, &rr->p.address) != 0))
        {
          buffer_free(&rdata);
          return -1;
        }
      } break;
      case TYPE_AAAA:
      {
        if(unlikely(buffer_put_addr6(&rdata, &rr->p.address) != 0))
        {
          buffer_free(&rdata);
          return -1;
        }
      } break;
      case TYPE_CNAME:
      {
        if(unlikely(put_name(client, &rdata, &rr->p.cname, (int)(client->packet.length + 2 + rdata.length)) != 0))
        {
          buffer_free(&rdata);
          return -1;
        }
      } break;
      case TYPE_MX:
      {
        if(unlikely(buffer_put16(&rdata, rr->p.mx.preference) != 0 || put_name(client, &rdata, &rr->p.mx.exchange, (int)(client->packet.length + 2 + rdata.length)) != 0))
        {
          buffer_free(&rdata);
          return -1;
        }
      } break;
      case TYPE_NS:
      {
        if(unlikely(put_name(client, &rdata, &rr->p.ns, (int)(client->packet.length + 2 + rdata.length)) != 0))
        {
          buffer_free(&rdata);
          return -1;
        }
      } break;
      case TYPE_PTR:
      {
        if(unlikely(put_name(client, &rdata, &rr->p.ptr, (int)(client->packet.length + 2 + rdata.length)) != 0))
        {
          buffer_free(&rdata);
          return -1;
        }
     } break;
      case TYPE_SOA:
      {
        if(unlikely(put_name(client, &rdata, &rr->p.soa.mname, (int)(client->packet.length + 2 + rdata.length)) != 0 ||
           put_name(client, &rdata, &rr->p.soa.rname, (int)(client->packet.length + 2 + rdata.length)) != 0 ||
           buffer_put32(&rdata, rr->p.soa.serial) != 0 || buffer_put32(&rdata, rr->p.soa.refresh) != 0 ||
           buffer_put32(&rdata, rr->p.soa.retry) != 0 || buffer_put32(&rdata, rr->p.soa.expire) != 0 ||
           buffer_put32(&rdata, rr->p.soa.minimum) != 0))
        {
          buffer_free(&rdata);
          return -1;
        }
      } break;
      case TYPE_SPF:
      {
        if(unlikely(buffer_put_raw(&rdata, rr->p.spf.value, rr->p.spf.length) != 0))
        {
          buffer_free(&rdata);
          return -1;
        }
      } break;
      case TYPE_TXT:
      {
        if(unlikely(buffer_put_raw(&rdata, rr->p.txt.value, rr->p.txt.length) != 0))
        {
          buffer_free(&rdata);
          return -1;
        }
      } break;
      default: break;
    }
  }
  else if(likely(rr->rclass == CLASS_CH && rr->rtype == TYPE_TXT))
  {
    if(unlikely(buffer_put_raw(&rdata, rr->p.txt.value, rr->p.txt.length) != 0))
    {
      buffer_free(&rdata);
      return -1;
    }
  }
  ret = buffer_put_buffer(&client->packet, &rdata);
  buffer_free(&rdata);
  return ret;
}

int write_packet(client_t *client)
{
  int i;
  if(unlikely(client == NULL)) return -1;
  client->packet.offset = 0;
  client->packet.length = 2;
  if(unlikely(put_header(client) != 0)) return -1;
  for(i = 0; i < (int)client->header.qdcount; i++) if(unlikely(put_question(client, i) != 0))  return -1;
  for(i = 0; i < (int)client->header.ancount; i++) if(unlikely(put_rr(client, RR_AN, i) != 0)) return -1;
  for(i = 0; i < (int)client->header.nscount; i++) if(unlikely(put_rr(client, RR_NS, i) != 0)) return -1;
  for(i = 0; i < (int)client->header.arcount; i++) if(unlikely(put_rr(client, RR_AR, i) != 0)) return -1;
  client->packet.value[0] = ((client->packet.length - 2) >> 8) & 0xff;
  client->packet.value[1] = (client->packet.length - 2) & 0xff;
  return 0;
}

int init_client(state_t *state, client_t *client, uint16_t max_udp_size)
{
  if(unlikely(state == NULL || client == NULL)) return -1;
  client->socket = INVALID_SOCKET;
  if(unlikely(max_udp_size < 512)) max_udp_size = 512;
  client->max_udp_size = max_udp_size;
  client->state = state;
  return buffer_resize(&client->packet, (uint32_t)max_udp_size + PACKET_LENGTH_SIZE);
}

void* free_client(client_t *client, int dynamic)
{
  int i;
  if(unlikely(client == NULL)) return NULL;
  if(likely(client->socket != INVALID_SOCKET))
  {
    (void)closesocket(client->socket);
    client->socket = INVALID_SOCKET;
  }
  buffer_free(&client->packet);
  for(i = 0; i < (int)client->header.qdcount; i++) free_question(client, i);
  for(i = 0; i < (int)client->header.ancount; i++) free_client_rr(client, RR_AN, i);
  for(i = 0; i < (int)client->header.nscount; i++) free_client_rr(client, RR_NS, i);
  for(i = 0; i < (int)client->header.arcount; i++) free_client_rr(client, RR_AR, i);
  if(dynamic != 0)
  {
    free(client);
    return NULL;
  }
  return client;
}

__attribute__((pure)) unsigned int allow_forward(client_t *client)
{
  int i;
  state_t *state;
  if(unlikely(client == NULL)) return 0;
  state = client->state;
  for(i = 0; i < state->allow_forwards_count; i++)
  {
    if(is_in_range(&client->from_addr, &state->allow_forwards[i]) != 0) return 1;
  }
  return 0;
}

uint32_t current_serial(void)
{
  #if !defined(WIN32) || defined(__CYGWIN__)
    struct tm t;
    time_t now = time(NULL);
    (void)gmtime_r(&now, &t);
    return (uint32_t)((t.tm_year + 1900) * 1000000 + (t.tm_mon + 1) * 10000 + t.tm_mday * 100 + 1);
  #else
    struct tm *t;
    time_t now = time(NULL);
    t = gmtime(&now); 
    return (uint32_t)((t->tm_year + 1900) * 1000000 + (t->tm_mon + 1) * 10000 + t->tm_mday * 100 + 1);
  #endif
}

double microtime(void)
{
  struct timeval tv = { 0, 0 };
  (void)gettimeofday(&tv, NULL);
  return (double)tv.tv_sec + (double)tv.tv_usec / 1000000.0;
}

__attribute__((const)) const char* type_name(uint16_t type)
{
  const char *ret;
  switch(type)
  {
    case TYPE_A:                ret = "A";                break;
    case TYPE_CNAME:            ret = "CNAME";            break;
    case TYPE_MX:               ret = "MX";               break;
    case TYPE_NS:               ret = "NS";               break;
    case TYPE_PTR:              ret = "PTR";              break;
    case TYPE_SOA:              ret = "SOA";              break;
    case TYPE_TXT:              ret = "TXT";              break;
    case TYPE_AAAA:             ret = "AAAA";             break;
    case TYPE_SPF:              ret = "SPF";              break;
    case TYPE_ANY:              ret = "ANY";              break;
    case TYPE_HINFO:            ret = "HINFO";            break;
    case TYPE_MD:               ret = "MD";               break;
    case TYPE_MF:               ret = "MF";               break;
    case TYPE_MB:               ret = "MB";               break;
    case TYPE_MG:               ret = "MG";               break;
    case TYPE_MR:               ret = "MR";               break;
    case TYPE_NULL:             ret = "NULL";             break;
    case TYPE_WKS:              ret = "WKS";              break;
    case TYPE_MINFO:            ret = "MINFO";            break;
    case TYPE_RP:               ret = "RP";               break;
    case TYPE_AFSDB:            ret = "AFSDB";            break;
    case TYPE_X25:              ret = "X25";              break;
    case TYPE_ISDN:             ret = "ISDN";             break;
    case TYPE_RT:               ret = "RT";               break;
    case TYPE_NSAP:             ret = "NSAP";             break;
    case TYPE_NSAPPTR:          ret = "NSAPPTR";          break;
    case TYPE_SIG:              ret = "SIG";              break;
    case TYPE_KEY:              ret = "KEY";              break;
    case TYPE_PX:               ret = "PX";               break;
    case TYPE_GPOS:             ret = "GPOS";             break;
    case TYPE_LOC:              ret = "LOC";              break;
    case TYPE_NXT:              ret = "NXT";              break;
    case TYPE_EID:              ret = "EID";              break;
    case TYPE_NIMLOC:           ret = "NIMLOC";           break;
    case TYPE_SRV:              ret = "SRV";              break;
    case TYPE_ATMA:             ret = "ATMA";             break;
    case TYPE_NAPTR:            ret = "NAPTR";            break;
    case TYPE_KX:               ret = "KX";               break;
    case TYPE_CERT:             ret = "CERT";             break;
    case TYPE_A6:               ret = "A6";               break;
    case TYPE_DNAME:            ret = "DNAME";            break;
    case TYPE_SINK:             ret = "SINK";             break;
    case TYPE_OPT:              ret = "OPT";              break;
    case TYPE_APL:              ret = "APL";              break;
    case TYPE_DS:               ret = "DS";               break;
    case TYPE_SSHFP:            ret = "SSHFP";            break;
    case TYPE_IPSECKEY:         ret = "IPSECKEY";         break;
    case TYPE_RRSIG:            ret = "RRSIG";            break;
    case TYPE_NSEC:             ret = "NSEC";             break;
    case TYPE_DNSKEY:           ret = "DNSKEY";           break;
    case TYPE_DHCID:            ret = "DHCID";            break;
    case TYPE_NSEC3:            ret = "NSEC3";            break;
    case TYPE_NSEC3PARAM:       ret = "NSEC3PARAM";       break;
    case TYPE_HIP:              ret = "HIP";              break;
    case TYPE_NINFO:            ret = "NINFO";            break;
    case TYPE_RKEY:             ret = "RKEY";             break;
    case TYPE_TALINK:           ret = "TALINK";           break;
    case TYPE_CDS:              ret = "CDS";              break;
    case TYPE_UINFO:            ret = "UINFO";            break;
    case TYPE_UID:              ret = "UID";              break;
    case TYPE_GID:              ret = "GID";              break;
    case TYPE_UNSPEC:           ret = "UNSPEC";           break;
    case TYPE_TKEY:             ret = "TKEY";             break;
    case TYPE_TSIG:             ret = "TSIG";             break;
    case TYPE_IXFR:             ret = "IXFR";             break;
    case TYPE_AXFR:             ret = "AXFR";             break;
    case TYPE_MAILB:            ret = "MAILB";            break;
    case TYPE_MAILA:            ret = "MAILA";            break;
    case TYPE_URI:              ret = "URI";              break;
    case TYPE_CAA:              ret = "CAA";              break;
    case TYPE_TA:               ret = "TA";               break;
    case TYPE_DLV:              ret = "DLV";              break;
    default:                    ret = "unknown";
  }
  return ret;
}

__attribute__((const)) const char* class_name(uint16_t dclass)
{
  const char *ret;
  switch(dclass)
  {
    case CLASS_IN:              ret = "IN";               break;
    case CLASS_CH:              ret = "CH";               break;
    case CLASS_HESIOD:          ret = "HESIOD";           break;
    case CLASS_NONE:            ret = "NONE";             break;
    case CLASS_ANY:             ret = "ANY";              break;
    default:                    ret = "unknown";
  }
  return ret;
}

__attribute__((const)) const char *answer_type_name(answer_t a)
{
  const char *ret;
  switch(a)
  {
    case ANSWER_UNSATISFIED:    ret = "unsatisfied";      break;
    case ANSWER_SATISFIED:      ret = "satisfied";        break;
    case ANSWER_BLOCKED:        ret = "blocked";          break;
    case ANSWER_NOTFOUND:       ret = "not found";        break;
    case ANSWER_NAMEERROR:      ret = "name error";       break;
    default:                    ret = "unknown";
  }
  return ret;
}

__attribute__((const)) const char *rcode_name(uint8_t rcode)
{
  const char *ret;
  switch(rcode)
  {
    case RCODE_NO_ERROR:        ret = "NO ERROR";         break;
    case RCODE_FORMAT_ERROR:    ret = "FORMAT ERROR";     break;
    case RCODE_SERVER_FAILURE:  ret = "SERVER FAILURE";   break;
    case RCODE_NAME_ERROR:      ret = "NAME ERROR";       break;
    case RCODE_NOT_IMPLEMENTED: ret = "NOT IMPLEMENTED";  break;
    case RCODE_REFUSED:         ret = "REFUSED";          break;
    case RCODE_NOT_FOUND:       ret = "NOT FOUND";        break;
    default:                    ret = "unknown";
  }
  return ret;
}

answer_t answer_question(client_t *client, rr_section_t rr_section, const buffer_t *name, uint16_t type, uint16_t dclass, int lazy)
{
  answer_t rc;
  state_t *state;
  int i;
  if(unlikely(client == NULL || name == NULL)) return ANSWER_UNSATISFIED;
  state = client->state;
  for(i = 0; i < state->module_count; i++)
  {
    if(likely(state->modules[i].answer != NULL))
    {
      rc = state->modules[i].answer(i, client, rr_section, name, type, dclass, lazy);
      DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s %s %s %s => %s", state->modules[i].name, (char*)name->value, class_name(dclass), type_name(type), answer_type_name(rc));
      if(rc != ANSWER_UNSATISFIED) return rc;
    }
  }
  return ANSWER_UNSATISFIED;
}

__attribute__((const)) int drop_priv(state_t *state)
{
  #if defined(WIN32) && !defined(__CYGWIN__)
    if(unlikely(state == NULL)) return -1;
    return 0;
  #else
    char *buf = NULL;
    struct passwd pwd, *result = NULL;
    const char *user, *chroot_path;
    if(unlikely(state == NULL)) return -1;
    user = config_get_const(state, "unnamed.user");
    chroot_path = config_get_const(state, "unnamed.chroot");
    memset(&pwd, 0, sizeof(pwd));
    if(user != NULL)
    {
      int s;
      int bufsize = (int)sysconf(_SC_GETPW_R_SIZE_MAX);
      if(unlikely(bufsize == -1)) bufsize = 16384;
      if(unlikely((buf = (char*)malloc((size_t)bufsize)) == NULL))
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "malloc failed: %s", ERRNO_STRING);
        return -1;
      }
      buf[0] = '\0';
      s = getpwnam_r(user, &pwd, buf, (size_t)bufsize, &result);
      if(unlikely(result == NULL))
      {
        if(likely(s == 0))
        {
          ERROR_IF(state->trace) (void)lprintf(TRACE, "User %s not found", user);
        }
        else
        {
          errno = s;
          ERROR_IF(state->trace) (void)lprintf(TRACE, "getpwnam_r failed: %s", ERRNO_STRING);
        }
        free(buf);
        return -1;
      }
    }
    if(chroot_path != NULL)
    {
      if(unlikely(chdir(chroot_path) == -1))
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "chdir failed: %s", ERRNO_STRING);
      }
      else if(unlikely(chroot(chroot_path) == -1))
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "chroot failed: %s", ERRNO_STRING);
      }
    }
    if(user != NULL)
    {
      if(unlikely(setgid(pwd.pw_gid) == -1)) ERROR_IF(state->trace) (void)lprintf(TRACE, "setgid failed: %s", ERRNO_STRING);
      if(unlikely(setuid(pwd.pw_uid) == -1)) ERROR_IF(state->trace) (void)lprintf(TRACE, "setuid failed: %s", ERRNO_STRING);
      free(buf);
    }
    return 0;
  #endif
}

__attribute__((pure)) static int match_pattern(const char *pattern, const char *string)
{
  register const char *p = pattern, *s = string;
  for(; *p != '\0'; ++s, ++p)
  {
    if(unlikely(*s == '\0') && likely(*p != '*')) return -1;
    if(*p == '*')  /* match zero or more */
    {
      int matched;
      while(*++p == '*') {}
      if(unlikely(*p == '\0')) return 1;
      if(*p != '[' && *p != '\\' && *p != '?')
      {
        while(*s != *p && *s != '\0') s++;
        if(*s != *p) return 0;
      }
      while(*s != '\0')
      {
        if((matched = match_pattern(p, s++)) != 0) return matched;
      }
      return 0;
    }
    if(*p == '[')  /* match set */
    {
      register const char sl = (char)(unlikely(*s <= 'Z' && *s >= 'A') ? (*s | 32) : *s);
      int matched = 0, reverse;
      if(unlikely((reverse = ((p[1] == '!') ? 1 : 0)) == 1)) ++p;
      if(unlikely(p[1] == '-'))
      {
        if(sl == '-')
        {
          if(unlikely(reverse == 1)) return 0;
          ++p;
          while(!(*p == ']' && likely(p[-1] != '\\')) && *p != '\0') ++p;
          continue;
        }
        ++p;
      }
      while(*++p != '\0')
      {
        if(unlikely(*p == ']')) break;
        if(unlikely(*p == '\\'))
        {
          if(unlikely(*++p == '\0')) break;
          if(*p != sl) continue;
          if(unlikely(reverse == 1)) return 0;
          matched = 1;
          ++p;
          while(!(*p == ']' && likely(p[-1] != '\\')) && *p != '\0') ++p;
          break;
        }
        if(sl == *p)
        {
          if(unlikely(reverse == 1)) return 0;
          matched = 1;
          ++p;
          while(!(*p == ']' && likely(p[-1] != '\\')) && *p != '\0') ++p;
          break;
        }
        if(unlikely(*p == '-'))
        {
          if(unlikely(sl >= p[-1] && sl <= p[1]))
          {
            if(unlikely(reverse == 1)) return 0;
            matched = 1;
            p += 2;
            while(!(*p == ']' && likely(p[-1] != '\\')) && *p != '\0') ++p;
            break;
          }
          ++p;
        }
      }
      if(matched == reverse) return 0;
    }
    else if(*p != '?')
    {
      if(*p == '\\') ++p;  /* escape character */
      if(unlikely(*s <= 'Z' && *s >= 'A'))
      {
        if(unlikely((*s | ' ') != *p)) return 0;
      }
      else if(unlikely(*s != *p))
      {
        return 0;
      }
    }
  }
  return (*s == '\0') ? 1 : 0;
}

__attribute__((pure)) int unnamed_match(const char *pattern, const char *string)
{
  if(unlikely(pattern == NULL || string == NULL)) return 0;
  return (int)((unsigned int)(match_pattern(pattern, string) + 1) >> 1);
}

static void treap_free(treap_t *t)
{
  if(t != NULL)
  {
    treap_free(t->left);
    treap_free(t->right);
    free(t->key);
    free(t);
  }
}

__attribute__((pure)) static treap_t* treap_find(treap_t *p, const char *key)
{
  while(p != NULL)
  {
    int c = strcmp(key, p->key);
    if(c == 0) break;
    p = (c < 0) ? p->left : p->right;
  }
  return p;
}

__attribute__((pure)) static treap_t* treap_find_partial(treap_t *p, const char *key)
{
  while(p != NULL)
  {
    int i, c = 1;
    for(i = 0; key[i] != '\0'; i++)
    {
      if(key[i] < p->key[i])
      {
        c = -1;
        break;
      }
      if(key[i] > p->key[i]) break;
    }
    if(p->key[i] == '\0') break;
    p = (c < 0) ? p->left : p->right;
  }
  return p;
}

static treap_t* treap_rotate_right(treap_t *p)
{
  treap_t *s = p->left;
  p->left = s->right;
  s->right = p;
  return s;
}

static treap_t* treap_rotate_left(treap_t *p)
{
  treap_t *s = p->right;
  p->right = s->left;
  s->left = p;
  return s;
}

#if 0
static treap_t* treap_remove(treap_t *p, char *key)
{
  int c = strcmp(key, p->key);
  if(c < 0)
  {
    p->left = treap_remove(p->left, key);
  }
  else if(c > 0)
  {
    p->right = treap_remove(p->right, key);
  }
  else if(p->left == NULL && p->right == NULL)
  {
    free(p->key);
    free(p);
    p = NULL;
  }
  else if(p->right == NULL || (p->left != NULL && p->left->priority > p->right->priority))
  {
    p = treap_rotate_right(p);
    p->right = treap_remove(p->right, key);
  }
  else
  {
    p = treap_rotate_left(p);
    p->left = treap_remove(p->left, key);
  }
  return p;
}
#endif

static treap_t* treap_add(treap_t *p, char *key)
{
  if(p == NULL)
  {
    if((p = (treap_t*)malloc(sizeof(*p))) != NULL)
    {
      p->priority = (int)random();
      p->key = key;
      p->left = p->right = NULL;
    }
  }
  else if(strcmp(key, p->key) < 0)
  {
    p->left = treap_add(p->left, key);
    if(p->priority < p->left->priority) p = treap_rotate_right(p);
  }
  else
  {
    p->right = treap_add(p->right, key);
    if(p->priority < p->right->priority) p = treap_rotate_left(p);
  }
  return p;
}

void match_free(pattern_list_t *patterns)
{
  int i;
  if(unlikely(patterns == NULL)) return;
  for(i = 0; i < patterns->pattern_count; i++) free(patterns->patterns[i]);
  free(patterns->patterns);
  treap_free(patterns->contains);
  treap_free(patterns->begins);
  treap_free(patterns->ends);
  treap_free(patterns->whole);
  memset(patterns, 0, sizeof(*patterns));
}

int match_add(state_t *state, pattern_list_t *patterns, char *pattern)
{
  int i, j, star_count = 0;
  char *out = (char*)malloc(strlen(pattern) + 1);
  if(unlikely(out == NULL)) return -1;
  for(i = j = 0; pattern[i] != '\0'; i++)
  {
    if(pattern[i] == '*')
    {
      star_count++;
      out[j++] = pattern[i];
    }
    else if(pattern[i] == '[' || pattern[i] == '\\' || pattern[i] == '?')
    {
      star_count = 5;
    }
    else
    {
      out[j++] = pattern[i];
    }
  }
  out[j] = '\0';

  if(star_count == 2 && out[0] == '*' && out[j-1] == '*')
  {
    memmove(out, out+1, (size_t)j-1);
    out[j-2] = '\0';
    patterns->contains = treap_add(patterns->contains, out);
    DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s: contains %s", pattern, out);
  }
  else if(star_count == 1 && out[0] == '*')
  {
    int a, z;
    char temp = '\0';
    for(a = 0, z = j - 1; a < j / 2; z--)    /* reverse string, shifting to the left one character */
    {
      out[a++] = out[z];
      out[z] = temp;
      temp = out[a];
    }
    patterns->ends = treap_add(patterns->ends, out);
    DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s: ends %s", pattern, out);
  }
  else if(star_count == 1 && out[j-1] == '*')
  {
    out[j-1] = '\0';
    patterns->begins = treap_add(patterns->begins, out);
    DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s: begins %s", pattern, out);
  }
  else if(star_count == 0)
  {
    patterns->whole = treap_add(patterns->whole, out);
    DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s: equals %s", pattern, out);
  }
  else
  {
    if(unlikely(patterns->pattern_count % REALLOC_SIZE == 0))
    {
      char **temp = (char**)realloc(patterns->patterns, (size_t)(patterns->pattern_count + REALLOC_SIZE) * sizeof(*patterns->patterns));
      if(unlikely(temp == NULL))
      {
        free(out);
        return -1;
      }
      patterns->patterns = temp;
    }
    patterns->patterns[patterns->pattern_count++] = out;
    for(i = patterns->pattern_count - 2; i >= 0 && strcmp(patterns->patterns[i], patterns->patterns[i+1]) > 0; i--)
    {
      char *p = patterns->patterns[i];
      patterns->patterns[i] = patterns->patterns[i+1];
      patterns->patterns[i+1] = p;
    }
    DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s: pattern %s", pattern, out);
  }
  return 0;
}

int match_compile(state_t *state, char **list, int count, pattern_list_t *patterns)
{
  int i;
  if(unlikely(state == NULL || list == NULL || patterns == NULL)) return -1;
  for(i = 0; i < count; i++)
  {
    if(match_add(state, patterns, list[i]) != 0) return -1;
  }
  return 0;
}

int match_pattern_list(const pattern_list_t *patterns, const char *string)
{
  int k, length;
  char small_reversed[32], *reversed;

  if(unlikely(patterns == NULL || string == NULL)) return 0;

  if(treap_find(patterns->whole, string) != NULL) return 1;
  if(treap_find_partial(patterns->begins, string) != NULL) return 1;

  length = (int)strlen(string);
  if((reversed = ((length < (int)sizeof(small_reversed)) ? small_reversed : (char*)malloc((size_t)length + 1))) != NULL)
  {
    int i, j;
    for(i = length - 1, j = 0; i >= 0; i--, j++) reversed[i] = string[j];
    reversed[length] = '\0';
    if(treap_find_partial(patterns->ends, reversed) != NULL)
    {
      if(reversed != small_reversed) free(reversed);
      return 1;
    }
    if(reversed != small_reversed) free(reversed);
  }

  for(k = 0; k < length; k++) if(treap_find_partial(patterns->contains, string + k) != NULL) return 1;

  for(k = 0; k < patterns->pattern_count; k++) if(unnamed_match(patterns->patterns[k], string) != 0) return 1;

  return 0;
}

uint16_t random_id(state_t *state)
{
  uint16_t id = 0;
  struct timeval tv = { 0, 0 };
  #if !defined(WIN32) || defined(__CYGWIN__)
    if(state->random_fd >= 0)
    {
      if(read(state->random_fd, &id, sizeof(id)) <= 0)
      {
        id = (uint16_t)(((long)&id ^ (long)&state) & 0xffff);
      }
    }
  #else
    id = (uint16_t)(((long)&id ^ (long)&state) & 0xffff);
  #endif
  (void)gettimeofday(&tv, NULL);
  return (uint16_t)(random() ^ tv.tv_usec ^ id ^ (long)&id);
}

#ifdef USE_PTHREADS
  static pthread_mutex_t lprintf_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif
int lprintf(state_t *state, const char *file, const char *function, int line, const char *fmt, ...)
{
  int length, h, m, s;
  double now;
  va_list ap;
  if(unlikely(state == NULL || file == NULL || function == NULL || fmt == NULL)) return -1;

  now = microtime();

  s = (int)now;
  h = s / 3600;
  s -= h * 3600;
  h %= 24;
  m = s / 60;
  s -= m * 60;
  now = now - (double)((int)now - s);

  va_start(ap, fmt);
  #ifdef USE_PTHREADS
    (void)pthread_mutex_lock(&lprintf_mutex);
  #endif
  if(unlikely(state->lprintf_fp == NULL))
  {
    (void)printf("%02d:%02d:%02d.%04d %s:%s:%d ", h, m, s, (int)(10000.0 * (now - (int)now)), file, function, line);
    length = vprintf(fmt, ap);
    (void)printf("\n");
  }
  else
  {
    (void)fprintf(state->lprintf_fp, "%02d:%02d:%02d.%04d %s:%s:%d ", h, m, s, (int)(10000.0 * (now - (int)now)), file, function, line);
    length = vfprintf(state->lprintf_fp, fmt, ap);
    (void)fprintf(state->lprintf_fp, "\n");
  }
  #ifdef USE_PTHREADS
    (void)pthread_mutex_unlock(&lprintf_mutex);
  #endif
  va_end(ap);
  return length;
}
